@@ -1,8 +1,3 @@
-/*
- * IDENTIFICATION
- *	pgroonga.c
- */
-
 #include "pgroonga.h"
 
 #include <access/reloptions.h>
@@ -20,9 +15,11 @@
 #include <storage/lmgr.h>
 #include <utils/array.h>
 #include <utils/builtins.h>
+#include <utils/guc.h>
 #include <utils/lsyscache.h>
 #include <utils/selfuncs.h>
 #include <utils/snapmgr.h>
+#include <utils/syscache.h>
 #include <utils/timestamp.h>
 #include <utils/tqual.h>
 #include <utils/typcache.h>
@@ -33,6 +30,9 @@
 #include <math.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#ifndef WIN32
+#    include <unistd.h>
+#endif
 
 #ifdef WIN32
 typedef struct _stat pgrn_stat_buffer;
@@ -49,6 +49,38 @@ PG_MODULE_MAGIC;
 static bool PGrnIsLZ4Available;
 static relopt_kind PGrnReloptionKind;
 
+static int PGrnLogType;
+enum PGrnLogType {
+	PGRN_LOG_TYPE_FILE,
+	PGRN_LOG_TYPE_WINDOWS_EVENT_LOG,
+	PGRN_LOG_TYPE_POSTGRESQL
+};
+static struct config_enum_entry PGrnLogTypeEntries[] = {
+	{"file",              PGRN_LOG_TYPE_FILE,              false},
+	{"windows_event_log", PGRN_LOG_TYPE_WINDOWS_EVENT_LOG, false},
+	{"postgresql",        PGRN_LOG_TYPE_POSTGRESQL,        false},
+	{NULL,                PGRN_LOG_TYPE_FILE,              false}
+};
+
+static char *PGrnLogPath;
+
+static int PGrnLogLevel;
+static struct config_enum_entry PGrnLogLevelEntries[] = {
+	{"none",      GRN_LOG_NONE,    false},
+	{"emergency", GRN_LOG_EMERG,   false},
+	{"alert",     GRN_LOG_ALERT,   false},
+	{"critical",  GRN_LOG_CRIT,    false},
+	{"error",     GRN_LOG_ERROR,   false},
+	{"warning",   GRN_LOG_WARNING, false},
+	{"notice",    GRN_LOG_NOTICE,  false},
+	{"info",      GRN_LOG_INFO,    false},
+	{"debug",     GRN_LOG_DEBUG,   false},
+	{"dump",      GRN_LOG_DUMP,    false},
+	{NULL,        GRN_LOG_NONE,    false}
+};
+
+static int PGrnLockTimeout;
+
 typedef struct PGrnOptions
 {
 	int32 vl_len_;
@@ -123,6 +155,7 @@ static slist_head PGrnScanOpaques = SLIST_STATIC_INIT(PGrnScanOpaques);
 PG_FUNCTION_INFO_V1(pgroonga_score);
 PG_FUNCTION_INFO_V1(pgroonga_table_name);
 PG_FUNCTION_INFO_V1(pgroonga_command);
+PG_FUNCTION_INFO_V1(pgroonga_snippet_html);
 
 PG_FUNCTION_INFO_V1(pgroonga_contain_text);
 PG_FUNCTION_INFO_V1(pgroonga_contain_text_array);
@@ -190,17 +223,159 @@ PGrnGetEncoding(void)
 	}
 }
 
+static void
+PGrnPostgreSQLLoggerLog(grn_ctx *ctx, grn_log_level level,
+						const char *timestamp, const char *title,
+						const char *message, const char *location,
+						void *user_data)
+{
+	const char levelMarks[] = " EACewnid-";
+
+	if (location && location[0])
+	{
+		ereport(LOG,
+				(errmsg("pgroonga:log: %s|%c|%s %s %s",
+						timestamp, levelMarks[level], title,
+						message, location)));
+	}
+	else
+	{
+		ereport(LOG,
+				(errmsg("pgroonga:log: %s|%c|%s %s",
+						timestamp, levelMarks[level], title, message)));
+	}
+}
+
+static grn_logger PGrnPostgreSQLLogger = {
+	GRN_LOG_DEFAULT_LEVEL,
+	GRN_LOG_TIME | GRN_LOG_MESSAGE,
+	NULL,
+	PGrnPostgreSQLLoggerLog,
+	NULL,
+	NULL
+};
+
+static void
+PGrnLogTypeAssign(int new_value, void *extra)
+{
+	switch (new_value) {
+	case PGRN_LOG_TYPE_WINDOWS_EVENT_LOG:
+		grn_windows_event_logger_set(ctx, "PGroonga");
+		break;
+	case PGRN_LOG_TYPE_POSTGRESQL:
+		grn_logger_set(ctx, &PGrnPostgreSQLLogger);
+		break;
+	default:
+		grn_logger_set(ctx, NULL);
+		break;
+	}
+}
+
+static void
+PGrnLogPathAssign(const char *new_value, void *extra)
+{
+	if (new_value) {
+		if (strcmp(new_value, "none") == 0) {
+			grn_default_logger_set_path(NULL);
+		} else {
+			grn_default_logger_set_path(new_value);
+		}
+	} else {
+		grn_default_logger_set_path(PGrnLogBasename);
+	}
+}
+
+static void
+PGrnLogLevelAssign(int new_value, void *extra)
+{
+	grn_default_logger_set_max_level(new_value);
+}
+
+static void
+PGrnLockTimeoutAssign(int new_value, void *extra)
+{
+	grn_set_lock_timeout(new_value);
+}
+
+static void
+PGrnInitializeVariables(void)
+{
+	DefineCustomEnumVariable("pgroonga.log_type",
+							 "Log type for PGroonga.",
+							 "Available log types: "
+							 "[file, windows_event_log, postgresql]. "
+							 "The default is file.",
+							 &PGrnLogType,
+							 PGRN_LOG_TYPE_FILE,
+							 PGrnLogTypeEntries,
+							 PGC_USERSET,
+							 0,
+							 NULL,
+							 PGrnLogTypeAssign,
+							 NULL);
+
+	DefineCustomStringVariable("pgroonga.log_path",
+							   "Log path for PGroonga.",
+							   "The default is "
+							   "\"${PG_DATA}/" PGrnLogBasename "\". "
+							   "Use \"none\" to disable file output.",
+							   &PGrnLogPath,
+							   NULL,
+							   PGC_USERSET,
+							   0,
+							   NULL,
+							   PGrnLogPathAssign,
+							   NULL);
+
+	DefineCustomEnumVariable("pgroonga.log_level",
+							 "Log level for PGroonga.",
+							 "Available log levels: "
+							 "[none, emergency, alert, critical, "
+							 "error, warning, notice, info, debug, dump]. "
+							 "The default is notice.",
+							 &PGrnLogLevel,
+							 GRN_LOG_DEFAULT_LEVEL,
+							 PGrnLogLevelEntries,
+							 PGC_USERSET,
+							 0,
+							 NULL,
+							 PGrnLogLevelAssign,
+							 NULL);
+
+	DefineCustomIntVariable("pgroonga.lock_timeout",
+							"Try pgroonga.lock_timeout times "
+							"at 1 msec intervals to "
+							"get write lock in PGroonga.",
+							"The default is 10000000. "
+							"It means that PGroonga tries to get write lock "
+							"between about 2.7 hours.",
+							&PGrnLockTimeout,
+							grn_get_lock_timeout(),
+							0,
+							INT_MAX,
+							PGC_USERSET,
+							0,
+							NULL,
+							PGrnLockTimeoutAssign,
+							NULL);
+
+	EmitWarningsOnPlaceholders("pgroonga");
+}
+
 static void
 PGrnEnsureDatabase(void)
 {
+	char *databasePath;
 	char path[MAXPGPATH];
 	grn_obj	*db;
 	pgrn_stat_buffer file_status;
 
 	GRN_CTX_SET_ENCODING(ctx, PGrnGetEncoding());
+	databasePath = GetDatabasePath(MyDatabaseId, DEFAULTTABLESPACE_OID);
 	join_path_components(path,
-						 GetDatabasePath(MyDatabaseId, DEFAULTTABLESPACE_OID),
+						 databasePath,
 						 PGrnDatabaseBasename);
+	pfree(databasePath);
 
 	if (pgrn_stat(path, &file_status) == 0)
 	{
@@ -222,40 +397,6 @@ PGrnEnsureDatabase(void)
 	}
 }
 
-static void
-PGrnLoggerLog(grn_ctx *ctx,
-			  grn_log_level level,
-			  const char *timestamp,
-			  const char *title,
-			  const char *message,
-			  const char *location,
-			  void *userData)
-{
-	const char levelTags[] = " EACewnid-";
-
-	if (location && *location)
-	{
-		ereport(LOG,
-				(errmsg("pgroonga: %s|%c|%s %s %s",
-						timestamp, levelTags[level], title, message, location)));
-	}
-	else
-	{
-		ereport(LOG,
-				(errmsg("pgroonga: %s|%c|%s %s",
-						timestamp, levelTags[level], title, message)));
-	}
-}
-
-static grn_logger PGrnLogger = {
-	GRN_LOG_DEFAULT_LEVEL,
-	GRN_LOG_TIME | GRN_LOG_MESSAGE,
-	NULL,
-	PGrnLoggerLog,
-	NULL,
-	NULL
-};
-
 static void
 PGrnOnProcExit(int code, Datum arg)
 {
@@ -388,6 +529,8 @@ PGrnInitializeOptions(void)
 void
 _PG_init(void)
 {
+	PGrnInitializeVariables();
+
 	if (grn_init() != GRN_SUCCESS)
 		ereport(ERROR,
 				(errcode(ERRCODE_SYSTEM_ERROR),
@@ -397,8 +540,6 @@ _PG_init(void)
 				(errcode(ERRCODE_SYSTEM_ERROR),
 				 errmsg("pgroonga: failed to initialize Groonga context")));
 
-	grn_logger_set(ctx, &PGrnLogger);
-
 	on_proc_exit(PGrnOnProcExit, 0);
 
 	GRN_VOID_INIT(&buffer);
@@ -1137,6 +1278,7 @@ pgroonga_table_name(PG_FUNCTION_ARGS)
 	Datum indexNameDatum = PG_GETARG_DATUM(0);
 	Datum indexOidDatum;
 	Oid indexOid;
+	Oid fileNodeOid;
 	char tableName[GRN_TABLE_MAX_KEY_SIZE];
 	char *copiedTableName;
 
@@ -1150,9 +1292,25 @@ pgroonga_table_name(PG_FUNCTION_ARGS)
 	}
 	indexOid = DatumGetObjectId(indexOidDatum);
 
+	{
+		HeapTuple tuple;
+		tuple = SearchSysCache1(RELOID, indexOid);
+		if (!HeapTupleIsValid(tuple)) {
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_NAME),
+					 errmsg("failed to find file node ID from index name: <%s>",
+							DatumGetCString(indexNameDatum))));
+		}
+		{
+			Form_pg_class indexClass = (Form_pg_class) GETSTRUCT(tuple);
+			fileNodeOid = indexClass->relfilenode;
+			ReleaseSysCache(tuple);
+		}
+	}
+
 	snprintf(tableName, sizeof(tableName),
 			 PGrnSourcesTableNameFormat,
-			 indexOid);
+			 fileNodeOid);
 	copiedTableName = pstrdup(tableName);
 	PG_RETURN_CSTRING(copiedTableName);
 }
@@ -1203,6 +1361,135 @@ pgroonga_command(PG_FUNCTION_ARGS)
 	PG_RETURN_TEXT_P(result);
 }
 
+static grn_obj *
+PGrnSnipCreate(ArrayType *keywords)
+{
+	grn_obj *snip;
+	int flags = GRN_SNIP_SKIP_LEADING_SPACES;
+	unsigned int width = 200;
+	unsigned int maxNResults = 3;
+	const char *openTag = "<span class=\"keyword\">";
+	const char *closeTag = "</span>";
+	grn_snip_mapping *mapping = GRN_SNIP_MAPPING_HTML_ESCAPE;
+
+	snip = grn_snip_open(ctx, flags, width, maxNResults,
+						 openTag, strlen(openTag),
+						 closeTag, strlen(closeTag),
+						 mapping);
+	if (!snip)
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_OUT_OF_MEMORY),
+				 errmsg("pgroonga: "
+						"failed to allocate memory for generating snippet")));
+		return NULL;
+	}
+
+	grn_snip_set_normalizer(ctx, snip, GRN_NORMALIZER_AUTO);
+
+	{
+		int i, n;
+
+		n = ARR_DIMS(keywords)[0];
+		for (i = 1; i <= n; i++)
+		{
+			Datum keywordDatum;
+			text *keyword;
+			bool isNULL;
+
+			keywordDatum = array_ref(keywords, 1, &i, -1, -1, false,
+									 'i', &isNULL);
+			if (isNULL)
+				continue;
+
+			keyword = DatumGetTextPP(keywordDatum);
+			grn_snip_add_cond(ctx, snip,
+							  VARDATA_ANY(keyword),
+							  VARSIZE_ANY_EXHDR(keyword),
+							  NULL, 0, NULL, 0);
+		}
+	}
+
+	return snip;
+}
+
+static grn_rc
+PGrnSnipExec(grn_obj *snip, text *target, ArrayType **snippetArray)
+{
+	grn_rc rc;
+	unsigned int i, nResults, maxTaggedLength;
+	char *buffer;
+	Datum *snippets;
+	int	dims[1];
+	int	lbs[1];
+
+	rc = grn_snip_exec(ctx, snip,
+					   VARDATA_ANY(target),
+					   VARSIZE_ANY_EXHDR(target),
+					   &nResults, &maxTaggedLength);
+	if (rc != GRN_SUCCESS)
+	{
+		return rc;
+	}
+
+	if (nResults == 0)
+	{
+		*snippetArray = construct_empty_array(TEXTOID);
+		return GRN_SUCCESS;
+	}
+
+	buffer = palloc(sizeof(char) * maxTaggedLength);
+	snippets = palloc(sizeof(Datum) * nResults);
+	for (i = 0; i < nResults; i++)
+	{
+		grn_rc rc;
+		unsigned int snippetLength = 0;
+
+		rc = grn_snip_get_result(ctx, snip, i, buffer, &snippetLength);
+		if (rc != GRN_SUCCESS)
+		{
+			pfree(buffer);
+			return rc;
+		}
+		snippets[i] = PointerGetDatum(cstring_to_text_with_len(buffer,
+															   snippetLength));
+    }
+	pfree(buffer);
+
+	dims[0] = nResults;
+	lbs[0] = 1;
+
+	*snippetArray = construct_md_array(snippets, NULL,
+									   1, dims, lbs,
+									   TEXTOID, -1, false, 'i');
+	return GRN_SUCCESS;
+}
+
+/**
+ * pgroonga.snippet_html(target text, keywords text[]) : text[]
+ */
+Datum
+pgroonga_snippet_html(PG_FUNCTION_ARGS)
+{
+	text *target = PG_GETARG_TEXT_PP(0);
+	ArrayType *keywords = PG_GETARG_ARRAYTYPE_P(1);
+	grn_obj *snip;
+	grn_rc rc;
+	ArrayType *snippets;
+
+	snip = PGrnSnipCreate(keywords);
+	rc = PGrnSnipExec(snip, target, &snippets);
+	grn_obj_close(ctx, snip);
+
+	if (rc != GRN_SUCCESS) {
+		ereport(ERROR,
+				(errcode(PGrnRCToPgErrorCode(rc)),
+				 errmsg("pgroonga: failed to compute snippets")));
+	}
+
+	PG_RETURN_POINTER(snippets);
+}
+
 static grn_bool
 pgroonga_contain_raw(const char *text, unsigned int textSize,
 					 const char *subText, unsigned int subTextSize)
@@ -2508,6 +2795,8 @@ PGrnRemoveUnusedTables(void)
 		nameSize = grn_table_cursor_get_key(ctx, cursor, (void **)&name);
 		nameEnd = name + nameSize;
 		relationID = strtol(name + strlen(min), &nameEnd, 10);
+		if (nameEnd[0] == '.')
+			continue;
 		relation = RelationIdGetRelation(relationID);
 		if (relation)
 		{
@@ -2521,7 +2810,7 @@ PGrnRemoveUnusedTables(void)
 			grn_obj *table;
 
 			snprintf(tableName, sizeof(tableName),
-					 PGrnLexiconNameFormat, i, relationID);
+					 PGrnLexiconNameFormat, relationID, i);
 			table = grn_ctx_get(ctx, tableName, strlen(tableName));
 			if (!table)
 				break;
